"""Test params parsing."""

import pytest_check as check

from docstripy.lines_routines import remove_indent
from docstripy.parse_doc.parse_params import parse_params_all


def test_params() -> None:
    """Test params parsing."""
    with open("tests/files/test3.py", "r", encoding="utf-8") as file:
        lines = file.readlines()
    rest_doc1 = remove_indent(lines[18:29])
    params_rest = parse_params_all(rest_doc1, style="rest")
    np_doc1 = remove_indent(lines[55:68])
    params_np = parse_params_all(np_doc1, style="numpy")
    google_doc1 = remove_indent(lines[93:102])
    params_google = parse_params_all(google_doc1, style="google")
    expected_list = [
        {
            "name": "name1",
            "type": "List[bool]",
            "optional": False,
            "description": ["A list of bools.\n"],
            "default": "",
        },
        {
            "name": "name2",
            "type": "list | str",
            "optional": True,
            "description": [
                "A list or a string.\n",
                "This parameter is very important.\n",
                "\n",
                "Very important.\n",
            ],
            "default": "an empty list",
        },
        {
            "name": "name3",
            "type": "Dict[str, int]",
            "optional": True,
            "description": ["A dictionary.\n"],
            "default": '{"a": 1}',
        },
    ]
    check.equal(params_rest, expected_list)
    check.equal(params_np, expected_list)
    check.equal(params_google, expected_list)

    rest_doc2 = remove_indent(lines[40:41])
    params_dict_rest2 = parse_params_all(rest_doc2, style="rest")
    np_doc2 = remove_indent(lines[81:84])
    params_dict_np2 = parse_params_all(np_doc2, style="numpy")
    google_doc2 = remove_indent(lines[113:115])
    params_dict_google2 = parse_params_all(google_doc2, style="google")
    expected_dict = [
        {"name": "a", "type": "", "optional": False, "description": [], "default": ""},
    ]
    check.equal(params_dict_rest2, expected_dict)
    check.equal(params_dict_np2, expected_dict)
    check.equal(params_dict_google2, expected_dict)



"""Parser for Google style docstrings."""

import re
from typing import Dict, List

from docstripy.lines_routines import find_prefix, remove_indent


def parse_params(lines: List[str], section_name: str) -> List[dict]:
    """Parse parameter section from Google format.

    Parameters
    ----------
    lines : List[str]
        Lines of docstring.
    section_name : str, optional
        Name of the section, one of "param", "return", "attribute", "raises".
    """
    # Find indentation
    if len(lines) < 2:
        return []
    # Skip the first line
    lines = lines[1:]
    lines = remove_indent(lines)
    params_list: List[dict] = []
    for line in lines:
        if not line.startswith(" ") and line.strip() != "":
            split_dots = re.split(r":\s|\n", line, maxsplit=1)
            param_name_type_opt = split_dots[0].strip()
            if "(" in param_name_type_opt:
                split_parenthesis = param_name_type_opt.split("(", maxsplit=1)
                param_name = split_parenthesis[0].strip()
                type_opt = split_parenthesis[1].split(")")[0].strip()
                if ", optional" in type_opt:
                    type_p = type_opt[:-10].strip()
                    optional = True
                elif "optional" in type_opt:
                    type_p = type_opt[:-8].strip()
                    optional = True
                else:
                    type_p = type_opt.strip()
                    optional = False
            else:
                param_name = param_name_type_opt.strip()
                type_p = ""
                optional = False  # by default
            params_list.append({})
            params_list[-1]["name"] = param_name
            params_list[-1]["type"] = type_p
            params_list[-1]["optional"] = optional
            params_list[-1]["description"] = []
            if len(split_dots) > 1:
                description = split_dots[1].lstrip()
                # Indent the first line of the description like other lines
                params_list[-1]["description"] = [description]
        elif len(params_list) > 0:
            params_list[-1]["description"].extend(remove_indent([line]))
    if section_name in ("return", "yield"):
        # When param_dict["type"] == "", it means that the name of the
        # return is not specified and param_dict["name"] is actually
        # the type.
        for param_dict in params_list:
            if param_dict["type"] == "":
                type_p = param_dict["name"]
                param_dict["name"] = param_dict["type"]
                param_dict["type"] = type_p
    return params_list


def parse_sections_ranges(lines: List[str]) -> Dict:
    """Parse google sections of a docstring."""
    params_start, params_end = find_prefix(
        lines, ("Arg:", "Args:", "Param:", "Params:"), ("\n", " ")
    )
    raises_start, raises_end = find_prefix(lines, ("Raise:", "Raises:"), ("\n", " "))
    returns_start, returns_end = find_prefix(
        lines, ("Return:", "Returns:"), ("\n", " ")
    )
    yields_start, yields_end = find_prefix(lines, ("Yields:", "Yield:"), ("\n", " "))
    attr_start, attr_end = find_prefix(
        lines, ("Attributes:", "Attribute:"), ("\n", " ")
    )
    return {
        "_parameters": [params_start, params_end],
        "_raises": [raises_start, raises_end],
        "_returns": [returns_start, returns_end],
        "_yields": [yields_start, yields_end],
        "_attributes": [attr_start, attr_end],
    }


def parse_wild_sections_ranges(lines: List[str]) -> Dict:
    """Parse wild sections of a docstring from google format."""
    sec_ranges = {}
    known_sections = [
        "parameter",
        "return",
        "raise",
        "arg",
        "attribute",
        "yield",
    ]
    known_sections += [f"{name}s" for name in known_sections]  # variations with s
    for i, line in enumerate(lines):
        if is_define_section(line):
            section_name = line[:-2].strip()
            if section_name.lower() not in known_sections:
                sec_ranges[section_name] = [i]
    for section_name in sec_ranges:
        start, _ = find_prefix(lines, (section_name,), ())
        sec_ranges.update({f"{section_name}": [start, -1]})
    return sec_ranges


def is_define_section(line: str) -> bool:
    """Check if a line define a Google wild section or not."""
    if not line or not line.endswith(":\n"):
        return False
    ord_n = ord(line[0])
    if not 65 <= ord_n <= 90:
        return False
    short_line = line[:-2].strip()
    if " " in short_line:
        # Multiple words
        return False
    return True



"""Global docstring parsing functions."""

from typing import List, Tuple

from docstripy.file_parser import parse_ranges
from docstripy.lines_routines import (
    clean_leading_empty,
    clean_trailing_empty,
    remove_indent,
    remove_quotes,
)
from docstripy.parse_doc.parse_def import parse_def
from docstripy.parse_doc.parse_params import parse_params_all
from docstripy.parse_doc.postprocessing import postprocess_title_parse
from docstripy.parse_doc.section_ranges import parse_sections_ranges


def parse_docstring(lines: List[str]) -> Tuple[List[List[int]], List[dict]]:
    """Docstring parser."""
    ranges_docstr, ranges_def = parse_ranges(lines)
    sections_list = []
    for range_docstr in ranges_docstr:
        range_def = find_range_matching(range_docstr, ranges_def)
        lines_docstr = lines[range_docstr[0] : range_docstr[1]]
        lines_def = lines[range_def[0] : range_def[1]]
        sections = parse_all(lines_docstr)
        sections = merge_docstr_def(sections, lines_def)
        sections_list.append(sections)
    return ranges_docstr, sections_list


def parse_all(lines_docstr: List[str]) -> dict:
    r"""Parse the whole docstring.

    Parameters
    ----------
    lines_docstr : List[str]
        Lines of the docstring.

    Returns
    -------
    sections : Dict
        Parsed sections. Can contain the following keys:
        - _title (List[str])
        - _parameters (List[Dict])
        - _returns (List[Dict])
        - _attributes (List[Dict])
        - _raises (List[Dict])
        - any other section name found in the docstring (List[str])

    Examples
    --------
    One can parse a docstring as follows:
    {
        '_title': ['This is a title\n', '\n', 'This is a description\n'],
        '_parameters': [
            {
                'name': 'param1',
                'type': 'int',
                'optional': False,
                'description': ['Description of param1\n'],
            },
            {
                'name': 'param2',
                'type': 'str',
                'optional': True,
                'description': ['Description of param2\n'],
                'default': '""',
            },
        ],
        '_returns': [
            {
                'type': 'int',
                'description': ['Description of the return\n'],
            },
        ],
        'Example': ['Example of a section\n'],
    }
    """
    lines, escaped = remove_quotes(lines_docstr)
    sec_ranges, style = parse_sections_ranges(lines)
    # Preprocess sections
    sections: dict = {}
    sections = {
        sec_name: lines[inds[0] : inds[1]] for (sec_name, inds) in sec_ranges.items()
    }
    sections = {
        sec_name: clean_trailing_empty(lines) for (sec_name, lines) in sections.items()
    }
    sections = {
        sec_name: clean_leading_empty(lines) for (sec_name, lines) in sections.items()
    }
    sections = {
        sec_name: remove_indent(lines) for (sec_name, lines) in sections.items()
    }
    # Manage param, return, attribute, raises using parse_params_all
    section_names_for_parse_params_all = {
        "_parameters": "param",
        "_returns": "return",
        "_yields": "yield",
        "_attributes": "attribute",
        "_raises": "raises",
    }
    for section_unders, section_name in section_names_for_parse_params_all.items():
        if section_unders in sections:
            section = parse_params_all(
                sections[section_unders],
                style=style,
                section_name=section_name,
            )
            if section:
                sections[section_unders] = section

    # Manage wild sections : remove header
    for section_name in sections:
        if not section_name.startswith("_"):
            header_len = 1
            if style == "numpy" and sections[section_name][1].strip().startswith("---"):
                header_len = 2
            sections[section_name] = sections[section_name][header_len:]
            if style in ("rest", "google"):
                sections[section_name] = remove_indent(sections[section_name])
    # NOTE : not necessary to modify _title section
    sections["_title"] = postprocess_title_parse(sections["_title"])
    sections["_escaped"] = escaped
    return sections


def merge_docstr_def(
    sections_docstr: dict,
    lines_def: List[str],
) -> dict:
    """Merge docstring and definition ranges."""
    if not lines_def:
        return sections_docstr
    fn_name, rtypes, args = parse_def(lines_def)
    if "_title" not in sections_docstr or not clean_leading_empty(
        sections_docstr["_title"]
    ):
        sections_docstr["_title"] = [" ".join(fn_name.split("_")).capitalize() + ".\n"]
    if "_parameters" in sections_docstr and sections_docstr["_parameters"]:
        # NOTE: only add/update param info if already any is provided in the doc
        for param in sections_docstr["_parameters"]:
            for arg in args:
                if param["name"] in (arg["name"], arg["name"].lstrip("*")):
                    param_name = param["name"]
                    # Priority to the docstring
                    for key in arg:
                        if arg[key]:
                            param[key] = arg[key]
                    param["name"] = param_name
        for arg in args:
            if arg["name"].lstrip("*") not in [
                param["name"].lstrip("*") for param in sections_docstr["_parameters"]
            ]:
                sections_docstr["_parameters"].append(arg)
    if (
        "_returns" in sections_docstr
        and sections_docstr["_returns"]
        and len(rtypes) == len(sections_docstr["_returns"])
    ):
        # NOTE: only add/update return info if already any is provided in the doc
        for i, return_arg in enumerate(sections_docstr["_returns"]):
            if "type" not in return_arg or not return_arg["type"]:
                return_arg["type"] = rtypes[i]
    return sections_docstr


def find_range_matching(
    range_docstr: List[int],
    ranges_def: List[List[int]],
) -> List[int]:
    """Find the range in ranges_def that matches range_docstr."""
    for range_def in ranges_def:
        if 0 <= range_docstr[0] - range_def[1] <= 1:
            return range_def
    return [-1, -1]



"""Postprocessing functions for some sections."""

from typing import List


def postprocess_title_parse(lines: List[str]) -> List[str]:
    """Post-process the title section."""
    new_lines = lines.copy()
    if new_lines[0].strip():
        new_lines[0] = new_lines[0].strip(" ")
        new_lines[0] = new_lines[0][0].capitalize() + new_lines[0][1:]
        if (
            (len(new_lines) >= 3 and not new_lines[1].strip()) or len(new_lines) == 1
        ) and new_lines[0][-2] not in (".", "!", "?"):
            new_lines[-1] = new_lines[-1][:-1] + ".\n"
    return new_lines



"""Parser for Numpy style docstrings."""

import re
from typing import Dict, List

from docstripy.lines_routines import find_prefix, remove_indent


def parse_params(lines: List[str], section_name: str = "param") -> List[dict]:
    """Parse parameter section from Numpy format.

    Parameters
    ----------
    lines : List[str]
        Lines of docstring.
    section_name : str, optional
        Name of the section, one of "param", "return", "attribute", "raises".
    """
    # Skip the 2 first lines
    lines = lines[2:]
    params_list: List[dict] = []
    for line in lines:
        if not line.startswith(" ") and line.strip() != "":
            split_dots = re.split(r":\s|\n", line, maxsplit=1)
            param_name = split_dots[0].strip()
            type_p, optional = "", False  # by default
            if len(split_dots) > 1:
                type_p = split_dots[1].strip()
                if type_p.endswith(", optional"):
                    type_p = type_p[:-10].strip()
                    optional = True
                elif type_p.endswith("optional"):
                    type_p = type_p[:-8].strip()
                    optional = True
                else:
                    type_p = type_p.strip()
                    optional = False
            params_list.append({})
            params_list[-1]["name"] = param_name
            params_list[-1]["description"] = []
            params_list[-1]["type"] = type_p
            params_list[-1]["optional"] = optional
        elif len(params_list) > 0:
            params_list[-1]["description"].append(line)
    for param_dict in params_list:
        param_dict["description"] = remove_indent(param_dict["description"])
    if section_name in ("return", "yield"):
        # When param_dict["type"] == "", it means that the name of the
        # return is not specified and param_dict["name"] is actually
        # the type.
        for param_dict in params_list:
            if param_dict["type"] == "":
                type_p = param_dict["name"]
                param_dict["name"] = param_dict["type"]
                param_dict["type"] = type_p
    return params_list


def parse_sections_ranges(lines: List[str]) -> Dict:
    """Parse numpy sections of a docstring."""
    params_start, _ = find_prefix(lines, ("Parameters", "Parameter"), (), dash=True)
    raises_start, _ = find_prefix(lines, ("Raises", "Raise"), (), dash=True)
    returns_start, _ = find_prefix(lines, ("Returns", "Return"), (), dash=True)
    yields_start, _ = find_prefix(lines, ("Yields", "Yield"), (), dash=True)
    attr_start, _ = find_prefix(lines, ("Attributes", "Attribute"), (), dash=True)
    return {
        "_parameters": [params_start, -1],
        "_raises": [raises_start, -1],
        "_returns": [returns_start, -1],
        "_yields": [yields_start, -1],
        "_attributes": [attr_start, -1],
    }


def parse_wild_sections_ranges(lines: List[str]) -> Dict:
    """Parse wild sections of a docstring from numpy format."""
    sec_ranges = {}
    known_sections = [
        "parameter",
        "return",
        "raise",
        "arg",
        "attribute",
        "yield",
    ]
    known_sections += [f"{name}s" for name in known_sections]  # add 's'
    for i, line in enumerate(lines):
        if (
            not line.startswith(" ")
            and len(lines) > i + 1
            and lines[i + 1].startswith("---")
        ):
            section_name = line[:-1].strip()
            if section_name.lower() not in known_sections:
                sec_ranges[section_name] = [i]
    for section_name in sec_ranges:
        start, _ = find_prefix(lines, (section_name,), ())
        sec_ranges.update({f"{section_name}": [start, -1]})
    return sec_ranges



# type: ignore
"""A test module.

Contains some functions for testing the parser.
"""
from typing import Dict, List, Tuple


def myfunc_rest(  # pylint: disable=W0102
    name1: List[bool],  # noqa
    name2: list | str = "",
    name3: Dict[str, int] = {"a,b": 1},
):
    r"""A function.

    It does something not important but the title
    is in multiple line.

    :param name1: A list of bools.
    :param name2: A list or a string.

        This parameter is very important.
    :type name2: list | bool, optional
    :param name3: A dictionary.
    :type name3: Dict[str, int] optional
    :raises ValueError: When something happened
    :raises ValueError: When something else happened
    :attribute nothing:
    :return result: The first result of the function
    :rtype: int
    :rtype: list
    :rtype result3: int
    :yield: A number
    :rtype: int

    Example:
        >>> myfunc_rest([True])
        1, [], 1
        >>> myfunc_rest([True], "\n")
        0, [], 0

    Notes:
        * This always works
        * This is a note
    """  # noqa: D401 D406 D407
    if name1[0] and len(name2) == 0 and name3["a"] == 1:
        return 1, [], 1
    return 0, [], 0


def myfunc_numpy(  # pylint: disable=W0102,W9010
    name1: List[bool], name2: list | str = "", name3: Dict[str, int] = {"a,b": 1}
) -> Tuple[int, List, int]:
    r"""A function.

    It does something not important but the title
    is in multiple line.

    Parameters
    ----------
    name1 :
        A list of bools.
    name2 :
        A list or a string.

        This parameter is very important.
    name3 : Dict[str, int] optional
        A dictionary, default is {"a,b": 1}.

    Raises
    ------
    ValueError :
        When something happened
    ValueError :
        When something else happened

    Attributes
    ----------
    nothing

    Returns
    -------
    result : int
        The first result of the function
    list
    result3 : int

    Yields
    ------
    int :
        A number

    Example
    -------
    >>> myfunc_rest([True])
    1, [], 1
    >>> myfunc_rest([True], "\n")
    0, [], 0

    Notes
    -----
    * This always works
    * This is a note
    """  # noqa: D401
    if name1[0] and len(name2) == 0 and name3["a"] == 1:
        return 1, [], 1
    return 0, [], 0


def myfunc_google(  # pylint: disable=W0102,W9015,W9010
    name1: List[bool], name2: list | str = "", name3: Dict[str, int] = {"a,b": 1}
) -> Tuple[int, List, int]:
    r"""A function.

    It does something not important but the title
    is in multiple line.

    Args:
        name1: A list of bools.
        name2 (list | str, optional): A list or a string.

            This parameter is very important.
        name3 (Dict[str, int] optional): A dictionary, default is {"c": 1}.

    Raises:
        ValueError: When something happened
        ValueError: When something else happened

    Attributes:
        nothing

    Returns:
        result (int): The first result of the function
        list
        result3 (int)

    Yields:
        int: A number.

    Example:
        >>> myfunc_rest([True])
        1, [], 1
        >>> myfunc_rest([True], "\n")
        0, [], 0

    Notes:
        * This always works
        * This is a note
    """  # noqa: D406 D407 D401
    if name1[0] and len(name2) == 0 and name3["a"] == 1:
        return 1, [], 1
    return 0, [], 0



"""Main functions for parsing and building docstrings."""

import argparse
import os
import os.path as osp

from docstripy.build_doc.main_builder import build_docstring
from docstripy.difference import Diff
from docstripy.lines_routines import add_indent, find_indent
from docstripy.parse_doc.main_parser import parse_docstring


def write_file(
    in_path: str,
    out_path: str,
    *,
    overwrite: bool,
    docstr_config: dict,
) -> None:
    """Write new docstrings on a file."""
    with open(in_path, encoding="utf-8") as file:
        file_lines = file.readlines()
    try:
        range_docstrs, sections_list = parse_docstring(file_lines)
    except (IndexError, ValueError) as err:
        raise ValueError("Error found during docstring parsing.") from err
    new_lines = []
    for range_doc, sections in zip(range_docstrs, sections_list):
        try:
            indent_base = find_indent(file_lines[range_doc[0] : range_doc[1]])
            docstring = build_docstring(
                sections=sections,
                docstr_config=docstr_config,
                indent_base=indent_base,
            )
        except (IndexError, ValueError) as err:
            raise ValueError(
                f"Error found at lines {range_doc[0]}-{range_doc[1]} "
                "during docstring building."
            ) from err
        docstring = add_indent(docstring, indent_base)
        new_lines.append("".join(docstring))
    # Write in file
    diff = Diff(range_docstrs, new_lines)
    file_new_lines = diff.apply_diff(file_lines)
    out_path = in_path if overwrite else out_path
    if osp.dirname(out_path):
        os.makedirs(osp.dirname(out_path), exist_ok=True)
    with open(out_path, "w", encoding="utf-8") as file:
        file.writelines(file_new_lines)


def write_files_recursive(
    in_path: str,
    out_path: str,
    *,
    overwrite: bool,
    docstr_config: dict,
) -> None:
    """Write new docstrings on all files in a folder."""
    error_paths = []
    for dir_path, _, file_names in os.walk(in_path):
        for file_name in file_names:
            if file_name.endswith(".py"):
                file_path = osp.join(dir_path, file_name)
                rel_path = osp.relpath(file_path, in_path)
                file_out_path = osp.join(out_path, rel_path)
                try:
                    write_file(
                        in_path=file_path,
                        out_path=file_out_path,
                        overwrite=overwrite,
                        docstr_config=docstr_config,
                    )
                except (IndexError, ValueError):
                    error_paths.append(file_path)
    if error_paths:
        err_message = "Error when parsing file(s):\n"
        err_message += "\n".join(error_paths)
        err_message += "\nRun `docstripy` on individual files to get lines error."
        print(err_message)


def parse_args() -> dict:
    """Command line parser for docstripy."""
    parser = argparse.ArgumentParser()
    parser.add_argument("in_path", help="File path or root directory path.", type=str)
    parser.add_argument(
        "-s", "--style", help="Style of the docstring", type=str, default="numpy"
    )
    parser.add_argument(
        "-o",
        "--out_path",
        help="Output files dir",
        type=str,
        default="",
    )
    parser.add_argument(
        "-w",
        "--overwrite",
        help="Whether to overwrite the files or not",
        action="store_true",
    )
    parser.add_argument(
        "--indent", help="Base indentation size (2, 4, ...)", type=int, default=4
    )
    parser.add_argument(
        "-l",
        "--length",
        help="Maximum lenght of lines in the docstring. By default, no limit.",
        type=int,
        default=-1,
    )
    args = parser.parse_args()
    docstr_config = {
        "style": args.style,
        "max_len": args.length,
        "indent": args.indent,
    }
    if args.out_path and args.overwrite:
        raise ValueError(
            "Cannot use both `--out_path`/`-o` and `--overwrite`/`-w` options."
        )
    if not args.out_path and not args.overwrite:
        raise ValueError(
            "You must specify an output path with `--out_path`/`-o` "
            "or overwrite file(s) with `--overwrite`/`-w` instead."
        )
    return {
        "in_path": args.in_path,
        "out_path": args.out_path,
        "docstr_config": docstr_config,
        "overwrite": args.overwrite,
    }


def main() -> None:
    """Rewrite file(s) docstrings main function."""
    cli_args = parse_args()
    if osp.isfile(cli_args["in_path"]):
        write_file(**cli_args)
    write_files_recursive(**cli_args)


if __name__ == "__main__":
    main()



"""Classes and functions to manage difference between files."""
from typing import List, Optional


class Diff:
    """Difference class to store the difference between two versions of a file.

    Parameters
    ----------
    ranges : Optional[List[List[int]]], optional
        List of ranges of docstring line numbers [start, end]. Includes line at start,
        excludes line at end. By default empty list.
        Note: the ranges will be represented in descending order.
    lines : Optional[List[str]], optional
        List of lines to add to the file. By default empty list.
    """

    def __init__(self, ranges: Optional[List[List[int]]], lines: Optional[List[str]]):
        self.ranges: List[List[int]] = ranges if ranges is not None else []
        self.lines: List[str] = lines if lines is not None else []
        self.satenize_check()

    def satenize_check(self) -> None:
        """Check if the ranges are overlapping."""
        for i, range1 in enumerate(self.ranges):
            for range2 in self.ranges[i + 1 :]:
                if range1[0] <= range2[0] < range1[1]:
                    raise ValueError("Found overlapping ranges.")

    def order_lists(self) -> None:
        """Order the ranges in descending order and lines accordingly."""
        sort_indices = sorted(range(len(self.ranges)), key=lambda i: self.ranges[i][0])
        self.ranges = [self.ranges[i] for i in sort_indices[::-1]]
        self.lines = [self.lines[i] for i in sort_indices[::-1]]

    @staticmethod
    def split_line(line: str) -> List[str]:
        r"""Split a line in multiple lines using '\n'."""
        if line[-1] == "\n":
            line = line[:-1]
        return [single_line + "\n" for single_line in line.split("\n")]

    def apply_diff(self, lines: List[str]) -> List[str]:
        """Apply the difference to a list of lines."""
        self.order_lists()
        self.satenize_check()
        new_lines = lines.copy()
        for i, line in enumerate(self.lines):
            start, end = self.ranges[i]
            del new_lines[start:end]
            new_lines[start:start] = self.split_line(line)
        return new_lines



"""Test file writting."""

import pytest
import pytest_check as check

from docstripy.difference import Diff


def test_diff() -> None:
    """Test difference class."""
    file_path = "tests/files/test1.py"
    with open(file_path, "r", encoding="utf-8") as file:
        lines = file.readlines()

    range_docstr = [[8, 24], [32, 33]]
    text = [
        '    r"""New function."""\n',
        '    """Other new function.\n    On multiple lines.\n    """\n',
    ]
    new_lines = Diff(range_docstr, text).apply_diff(lines)

    check.equal(new_lines[8], '    r"""New function."""\n')
    check.equal(new_lines[9], "    if n < 0:\n")
    check.equal(new_lines[17], '    """Other new function.\n')
    check.equal(new_lines[18], "    On multiple lines.\n")
    check.equal(new_lines[19], '    """\n')

    new_lines = Diff(range_docstr[::-1], text[::-1]).apply_diff(lines)

    check.equal(new_lines[8], '    r"""New function."""\n')
    check.equal(new_lines[9], "    if n < 0:\n")
    check.equal(new_lines[17], '    """Other new function.\n')
    check.equal(new_lines[18], "    On multiple lines.\n")
    check.equal(new_lines[19], '    """\n')

    with pytest.raises(ValueError, match="Found overlapping ranges."):
        Diff([[7, 23], [15, 32]], text).apply_diff(lines)
